#include <QWidget>
#include <QVBoxLayout>
#include <QLineEdit>
#include <QPushButton>
#include <QTextEdit>
#include <QLabel>
#include <QString>
#include <QThread>
#include <QCheckBox>
#include <QHBoxLayout>
#include <cmath>
#include <QSpinBox>
#include <QApplication>

using namespace std;
using namespace std::chrono;

void countSort(vector<short>& inputArray)
{
    short min = inputArray[0];
    short max = inputArray[0];
    for (unsigned int i = 1; i < inputArray.size(); i++) {
        if (inputArray[i] < min) min = inputArray[i];
        if (inputArray[i] > max) max = inputArray[i];
    }
    unsigned int size = max - min + 1;
    unsigned int nums[size];
    for (unsigned int i=0;i<size;i++)nums[i]=0;
    for (unsigned int i = 0; i < inputArray.size(); i++)nums[inputArray[i] - min]++;
    unsigned int pos = 0;
    for (unsigned int i = 0; i < size; i++) {
        while (nums[i] > 0) {
            inputArray[pos] = i + min;
            pos++;
            nums[i]--;
        }
    }
}

class SortWorker : public QObject {
    Q_OBJECT
public:
    SortWorker(vector<short>& arr, unsigned int n, bool individualTask) 
        : arr(arr), n(n), individualTask(individualTask), shouldStop(false) {}
    
    void stop() {
        shouldStop = true;
    }
    
public slots:
    void performSort() {
        auto start_time = high_resolution_clock::now();
        countSort(arr);
        auto duration = duration_cast<microseconds>(high_resolution_clock::now() - start_time);
        emit progressUpdated(100, n, n);
        emit sortingFinished(duration.count(), is_sorted(arr.data(), n, [](int a, int b) { return a < b; }));
    }
    
signals:
    void progressUpdated(int progress, unsigned int current, unsigned int total);
    void sortingFinished(int duration, bool sorted);
    void sortingCancelled();
    
private:
    vector<short>& arr;
    unsigned int n;
    bool individualTask;
    std::atomic<bool> shouldStop;
    template<typename Compare>
    bool is_sorted(short* arr, unsigned int n, Compare comp) {
        for (unsigned int i = 0; i < n - 1; i++) {
            if (comp(arr[i+1], arr[i])) return false;
        }
        return true;
    }
};
class SortingApp : public QWidget {
    Q_OBJECT
public:
    SortingApp(QWidget *parent = nullptr) : QWidget(parent), worker(nullptr), isSorting(false) {
        setupUI();
    }

    ~SortingApp() {
        if (workerThread.isRunning()) {
            workerThread.quit();
            workerThread.wait();
        }
    }

private:
    QSpinBox *inputField;
    QPushButton *sortButton;
    QTextEdit *outputField;
    QCheckBox *showArrayCheckBox;
    vector<short> arr;
    QThread workerThread;
    SortWorker *worker;
    QCheckBox *individualTaskCheckBox;
    bool isSorting;

    void setupUI() {
        QVBoxLayout *mainLayout = new QVBoxLayout(this);

        QLabel *label = new QLabel("Скільки чисел сортувати?", this);
        mainLayout->addWidget(label);

        inputField = new QSpinBox(this);
        inputField->setRange(1, 10000000);
        inputField->setValue(10);  // Встановлюємо початкове значення
        mainLayout->addWidget(inputField);

        sortButton = new QPushButton("Сортувати", this);
        connect(sortButton, &QPushButton::clicked, this, &SortingApp::onSortButtonClicked);
        mainLayout->addWidget(sortButton);

        QHBoxLayout *checkBoxLayout = new QHBoxLayout();
        
        showArrayCheckBox = new QCheckBox("Показати масив до та після сортування", this);
        checkBoxLayout->addWidget(showArrayCheckBox);

        individualTaskCheckBox = new QCheckBox("Індивідуальне завдання", this);
        checkBoxLayout->addWidget(individualTaskCheckBox);

        mainLayout->addLayout(checkBoxLayout);

        outputField = new QTextEdit(this);
        outputField->setReadOnly(true);
        mainLayout->addWidget(outputField);

        setLayout(mainLayout);
        setWindowTitle("Сортування масиву");
        resize(500, 500);
    }

    void sortArray() {
        arr.clear();
        arr.resize(inputField->value());
        for (unsigned int i = 0; i < inputField->value(); i++) arr[i] = rand() % 20001 - 10000;
        outputField->clear();
        if (individualTaskCheckBox->isChecked()){
            outputField->append("==Режим індивідуального завдання #10==\n");
            short arr2[inputField->value()];
            for (unsigned int i = 0; i < inputField->value(); i++) arr2[i] = rand() % 20001 - 10000;
            if (showArrayCheckBox->isChecked())outputField->append("Перший масив: "+arrayToString(arr)+"\nДругий масив: "+arrayToString(arr2,inputField->value())+'\n');
            for (unsigned int i=0;i<arr.size();i++)if(!(arr[i]%2)){arr.erase(arr.begin()+i);i--;}
            for (unsigned int i=0;i<inputField->value();i++)if(arr[i]%2)arr.push_back(arr2[i]);
        }
        if (showArrayCheckBox->isChecked()) {
            outputField->append("Масив до сортування: "+arrayToString(arr));
        }
            worker = new SortWorker(arr, arr.size(), individualTaskCheckBox->isChecked());
            worker->moveToThread(&workerThread);
            connect(this, &SortingApp::startSort, worker, &SortWorker::performSort);
            connect(worker, &SortWorker::progressUpdated, this, &SortingApp::updateProgress);
            connect(worker, &SortWorker::sortingFinished, this, &SortingApp::finishSorting);
            connect(&workerThread, &QThread::finished, worker, &QObject::deleteLater);
        workerThread.start();
        emit startSort();
    }
    
    QString arrayToString(const vector<short>& arr) {
        QString result;
        for (size_t i = 0; i < arr.size(); ++i) {
            result += QString::number(static_cast<int>(arr[i]));
            if (i < arr.size() - 1) result += ", ";
        }
        return result;
    }
    QString arrayToString(const short *arr,unsigned int size) {
        QString result;
        for (size_t i = 0; i<size; ++i) {
            result += QString::number(static_cast<int>(arr[i]));
            if (i<size - 1) result += ", ";
        }
        return result;
    }
private slots:
    void onSortButtonClicked() {
        if (!isSorting) {
        isSorting = true;
        sortButton->setText("Скасувати");
            sortArray();
        } else resetSortingState();
    }

    void updateProgress(int progress, unsigned int current, unsigned int total) {
        outputField->append(QString("Відсортовано %1% (%2 з %3 елементів)").arg(progress).arg(current).arg(total));
    }

    void finishSorting(int duration, bool sorted) {
        outputField->append(QString("\nЧас сортування: %1 мікросекунд").arg(duration));
        outputField->append(sorted ? "Масив відсортовано правильно." : "Помилка: масив не відсортовано правильно.");
        
        if (showArrayCheckBox->isChecked()) {
            outputField->append("\nМасив після сортування:");
            outputField->append(arrayToString(arr));
            arr.clear();
        }
        
        resetSortingState();
    }
    void resetSortingState() {
        isSorting = false;
        sortButton->setText("Сортувати");
        if(worker){worker->disconnect();worker->deleteLater();}
        worker=nullptr;
    }

signals:
    void startSort();
};

int main(int argc, char *argv[]) {
    QApplication app(argc, argv);
    SortingApp sortingApp;
    sortingApp.show();
    return app.exec();
}
#include "main.moc"
